"""
Example random bots.
"""

# =============================================================================

import random
from typing import Dict

from alicechess.game_state import GameState
from alicechess.player import Player
from alicechess.position import Move
from alicechess.utils import PieceType, PromoteType

# =============================================================================

__all__ = (
    "RandomPlayer",
    "PrioritizedRandomPlayer",
)

# =============================================================================


class RandomPlayer(Player):
    """A bot that picks a random move to make."""

    def make_move(self, game_state: GameState) -> Move:
        return random.choice(list(game_state.yield_player_moves()))

    def promote(self, game_state: GameState) -> PromoteType:
        return PromoteType.by_index(random.randrange(len(PromoteType)))


class PrioritizedRandomPlayer(Player):
    """A bot that picks a random move to make, prioritizing moves that
    put the other king in checkmate or check, then prioritizing moves
    that capture an enemy piece. Out of those priorities, it tries to
    not put the moved piece in a threatened position, and also tries to
    capture more valuable pieces.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # maps: int -> promote type
        self._saved_promotions: Dict[int, PromoteType] = {}

    def make_move(self, game_state: GameState) -> Move:
        PIECE_VALUES = {
            PieceType.KING: 0,
            PieceType.QUEEN: 8,
            PieceType.ROOK: 5,
            PieceType.KNIGHT: 3,
            PieceType.BISHOP: 3,
            PieceType.PAWN: 1,
        }

        max_weight = None
        max_weight_moves = []
        for move in game_state.yield_player_moves():
            next_state = game_state.make_move(move)
            if next_state.needs_promotion():
                # try every promotion
                check_states = []
                for promote_type in PromoteType:
                    state = next_state.promote(promote_type)
                    check_states.append(state)
                    self._saved_promotions[state.id] = promote_type
            else:
                check_states = [next_state]

            for state in check_states:
                move_made = state.move
                is_threatened = move_made.piece_moved.is_threatened

                priority = 0
                if state.is_in_checkmate():
                    priority = 3
                elif state.is_in_check():
                    priority = 2
                elif move_made.move_captured:
                    priority = 1

                # not capturing a piece has value 0
                value = 0
                if move_made.move_captured:
                    value = PIECE_VALUES[move_made.piece_captured.type]

                weight = (priority, 0 if is_threatened else 1, value)
                if max_weight is None or weight > max_weight:
                    max_weight = weight
                    max_weight_moves = [move]
                elif weight == max_weight:
                    max_weight_moves.append(move)

        return random.choice(max_weight_moves)

    def promote(self, game_state: GameState) -> PromoteType:
        promote_type = self._saved_promotions.get(game_state.id, None)
        if promote_type is not None:
            return promote_type
        return PromoteType.by_index(random.randrange(len(PromoteType)))



"""
Position, BoardPosition, Move, PieceMove, and PieceMoved classes.
"""

# =============================================================================

from functools import total_ordering
from typing import Iterable, Optional, Self, Tuple

from alicechess.utils import check_brc

# =============================================================================

__all__ = (
    "Position",
    "BoardPosition",
    "Move",
    "PieceMove",
    "PieceMoved",
)

# =============================================================================


def _rc_code(r: int, c: int) -> str:
    row_number = str(8 - r)
    col_letter = "abcdefgh"[c]
    return f"{col_letter}{row_number}"


def _rc_from_code(code: str) -> Tuple[int, int]:
    if len(code) != 2:
        raise ValueError("`code` must be length 2")
    letter, number = code.lower()
    if not number.isdigit():
        raise ValueError(f"Invalid code: not a digit: {number!r}")
    try:
        c = "abcdefgh".index(letter)
    except ValueError:
        raise ValueError(
            f"Invalid code: letter must be a-h (got {letter!r})"
        ) from None
    r = 8 - int(number)
    if not 0 <= r < 8:
        raise ValueError(
            f"Invalid code: number must be within [1, 8] (got {number})"
        )
    return r, c


# =============================================================================


@total_ordering
class _ImmutableParts:
    """A base class for common functions and methods of an immutable
    object made up of "parts".

    The constructor must define the properties:
        _parts: Defines the parts. Must be iterable.
        _str: The __str__ representation.
        _args_tuple: The args as a tuple, as they should appear in the
            __repr__ representation.
    """

    # pylint: disable=no-member

    @classmethod
    def of(cls, obj) -> Self:
        if obj is None:
            return None
        if isinstance(obj, cls):
            return obj
        try:
            args = tuple(obj)
        except TypeError:
            pass
        else:
            # if iterable, use it as args
            return cls(*args)
        raise TypeError(
            f"Cannot create {cls.__name__} from {obj.__class__.__name__}"
        )

    @classmethod
    def to_str(cls, iterable: Iterable[Self], sep: str = " ") -> str:
        """Returns the objs in the given iterable as a string."""
        return sep.join(map(lambda obj: str(cls.of(obj)), iterable))

    def __str__(self):
        return self._str

    def __repr__(self):
        args_str = ", ".join(map(repr, self._args_tuple))
        return f"{self.__class__.__name__}({args_str})"

    def __hash__(self):
        return hash(self._parts)

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self._parts < other._parts
        return self._parts < other

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._parts == other._parts
        return self._parts == other

    def __iter__(self):
        return iter(self._parts)


# =============================================================================


class PositionBase(_ImmutableParts):
    """Position base class."""

    def __init__(self, pos, code):
        self._pos = pos
        self._code = code

        self._parts = self._pos
        self._str = self._code
        self._args_tuple = self._pos

    @classmethod
    def from_code(cls, code: str):
        raise NotImplementedError(
            f"{cls.__name__}: function `from_code()` not implemented"
        )

    @property
    def r(self) -> int:
        """The row."""
        return self._pos[-2]

    @property
    def c(self) -> int:
        """The column."""
        return self._pos[-1]

    @property
    def code(self) -> str:
        """The position code (in algebraic notation)."""
        return self._code


class Position(PositionBase):
    """Represents a position on a single board.

    Properties:
        pos (Tuple[int, int]): The row and column.
        r (int): The row.
        c (int): The column.
        code (str): The position code (in algebraic notation).
    """

    def __init__(self, r, c):
        """Initializes a Position."""
        check_brc(r=r, c=c)
        pos = (r, c)
        code = _rc_code(r, c)
        super().__init__(pos, code)

    @classmethod
    def from_code(cls, code: str) -> "Position":
        """Constructs a Position based on its code.

        Args:
            code (str): The code in algebraic notation (of length 2).

        Raises:
            ValueError: If `code` is invalid.
        """
        r, c = _rc_from_code(code)
        return cls(r, c)

    @property
    def pos(self) -> Tuple[int, int]:
        """The row and column."""
        return self._pos


class BoardPosition(PositionBase):
    """Represents a position on either board.

    Properties:
        pos (Tuple[int, int, int]): The board number, row, and column.
        bn (int): The board number.
        r (int): The row.
        c (int): The column.
        code (str): The position code (in algebraic notation).
    """

    def __init__(self, bn, r, c):
        """Initializes a BoardPosition."""
        check_brc(bn=bn, r=r, c=c)
        pos = (bn, r, c)
        code = "AB"[bn] + _rc_code(r, c)
        super().__init__(pos, code)

    @classmethod
    def from_code(cls, code: str) -> "BoardPosition":
        """Constructs a BoardPosition based on its code.

        Args:
            code (str): The code in algebraic notation (of length 3).

        Raises:
            ValueError: If `code` is invalid.
        """
        if len(code) != 3:
            raise ValueError("`code` must have length 3")
        board = code[0].upper()
        try:
            bn = "AB".index(board)
        except ValueError:
            raise ValueError(
                f"Invalid code: board must be A or B (got {board!r})"
            ) from None
        r, c = _rc_from_code(code[1:])
        return cls(bn, r, c)

    @property
    def pos(self) -> Tuple[int, int, int]:
        """The board number, row, and column."""
        return self._pos

    @property
    def bn(self) -> int:
        """The board number."""
        return self._pos[0]


# =============================================================================


class Move(_ImmutableParts):
    """Represents a move of a piece.

    Properties:
        pos (BoardPosition): The piece position.
        target (Position): The target position.
        capture_pos (BoardPosition): The capture position.
        result_pos (BoardPosition): The piece's resulting position,
            after teleporting.
    """

    def __init__(self, pos: BoardPosition, target: Position):
        """Initializes a Move."""
        self._pos = BoardPosition.of(pos)
        self._target = Position.of(target)
        self._capture_pos = BoardPosition(self._pos.bn, *self._target)
        self._result_pos = BoardPosition(1 - self._pos.bn, *self._target)

        self._parts = (self._pos, self._target)
        self._str = f"{self._pos}-{self._target}"
        self._args_tuple = self._parts

    @property
    def pos(self) -> BoardPosition:
        """The piece position."""
        return self._pos

    @property
    def target(self) -> Position:
        """The target position."""
        return self._target

    @property
    def capture_pos(self) -> BoardPosition:
        """The capture position.

        This is `target` with the board number.
        """
        return self._capture_pos

    @property
    def result_pos(self) -> BoardPosition:
        """The piece's resulting position, after teleporting."""
        return self._result_pos


class PieceMove(Move):
    """Represents a possible move for a piece, which is a Move that
    includes information about the piece.

    Properties:
        pos (BoardPosition): The piece position.
        target (Position): The target position.
        capture_pos (BoardPosition): The capture position.
        result_pos (BoardPosition): The piece's resulting position,
            after teleporting.
        piece_moved (Piece): The piece that is being moved.
    """

    def __init__(self, pos: BoardPosition, target: Position, piece: "Piece"):
        """Initializes a PieceMove."""
        super().__init__(pos, target)

        self._piece_moved = piece

        self._args_tuple = (self._pos, self._target, self._piece_moved)

    @property
    def piece_moved(self) -> "Piece":
        """The piece that was moved."""
        return self._piece_moved


class PieceMoved(PieceMove):
    """Represents a move done by a piece, which is a PieceMove that
    includes information about a possible capture.

    Properties:
        pos (BoardPosition): The piece position.
        target (Position): The target position.
        capture_pos (BoardPosition): The capture position.
        result_pos (BoardPosition): The piece's resulting position,
            after teleporting.
        piece_moved (Piece): The piece that was moved.
        move_captured (bool): Whether the move captured a piece.
        piece_captured (Optional[Piece]): The piece that was captured,
            if given.
    """

    def __init__(
        self,
        pos: BoardPosition,
        target: Position,
        piece: "Piece",
        captured: Optional["Piece"],
    ):
        """Initializes a PieceMoved."""
        super().__init__(pos, target, piece)

        self._piece_moved = piece
        self._move_captured = captured is not None
        self._piece_captured = captured

        self._args_tuple = (
            self._pos,
            self._target,
            self._piece_moved,
            self._piece_captured,
        )

    @property
    def move_captured(self) -> bool:
        """Whether the move captured a piece."""
        return self._move_captured

    @property
    def piece_captured(self) -> Optional["Piece"]:
        """The piece that was captured, if any."""
        return self._piece_captured



"""
Starts and runs a game in a window between two human players.
"""

# ======================================================================

import sys

from alicechess import bots
from alicechess.game import Game
from alicechess.player import HumanPlayer

# ======================================================================


def main():
    _, *args = sys.argv

    if len(args) == 0:
        white = HumanPlayer
        black = HumanPlayer
    elif len(args) == 2:
        pass

    Game(white=HumanPlayer, black=HumanPlayer).start_window()


if __name__ == "__main__":
    main()



"""
Pieces submodule.
"""

# =============================================================================

import typing as _t

from alicechess.pieces.king import King
from alicechess.pieces.pawn import Pawn
from alicechess.pieces.piece import Piece
from alicechess.utils import Color, PieceType, PromoteType

# =============================================================================

__all__ = (
    "Piece",
    "King",
    "Queen",
    "Rook",
    "Knight",
    "Bishop",
    "Pawn",
    "make_promoted",
)

# =============================================================================


def _make_simple_piece(piece_type, rows, columns=None) -> _t.Type[Piece]:
    """Makes a piece."""
    # pylint: disable=missing-class-docstring

    class Subclass(Piece):
        _type = piece_type

        def _check_start_position(self) -> bool:
            if self._pos.bn != 0:
                return False
            if self._color is Color.WHITE:
                row = rows[0]
            else:
                row = rows[1]
            if self._pos.r != row:
                return False
            if columns is not None and self._pos.c not in columns:
                return False
            return True

    piece_name = piece_type.title()
    Subclass.__name__ = piece_name
    Subclass.__doc__ = f"{piece_name} piece."

    return Subclass


# queen is always on the left for this particular board setup
Queen = _make_simple_piece(PieceType.QUEEN, (7, 0), (3,))
Rook = _make_simple_piece(PieceType.ROOK, (7, 0), (0, 7))
Knight = _make_simple_piece(PieceType.KNIGHT, (7, 0), (1, 6))
Bishop = _make_simple_piece(PieceType.BISHOP, (7, 0), (2, 5))

# =============================================================================


def make_promoted(
    pawn: Pawn, promote_type: PromoteType
) -> _t.Union[Queen, Rook, Knight, Bishop]:
    PROMOTE_CLASSES = {
        PromoteType.QUEEN: Queen,
        PromoteType.ROOK: Rook,
        PromoteType.KNIGHT: Knight,
        PromoteType.BISHOP: Bishop,
    }
    return PROMOTE_CLASSES[promote_type](
        pawn.id, pawn.color, pawn.pos, has_moved=pawn.has_moved
    )



"""
Pawn piece.
"""

# =============================================================================

from alicechess.pieces.piece import Piece
from alicechess.utils import Color, PieceType

# =============================================================================


class Pawn(Piece):
    """Pawn piece.

    Properties:
        dr (int): The direction of movement.
        can_promote (bool): Whether the pawn can be promoted.
    """

    _type = PieceType.PAWN

    def __init__(self, *args, **kwargs):
        """Initializes a Pawn piece."""
        super().__init__(*args, **kwargs)

        if self._color is Color.WHITE:
            # going up
            self._dr = -1
            self._start_row = 6
            self._promote_row = 0
        else:
            # going down
            self._dr = 1
            self._start_row = 1
            self._promote_row = 7

    def _check_start_position(self) -> bool:
        return self._pos.bn == 0 and self._pos.r == self._start_row

    @property
    def dr(self) -> int:
        """The direction of movement."""
        return self._dr

    @property
    def can_promote(self) -> bool:
        """Whether the pawn can be promoted."""
        return self._pos.r == self._promote_row

    def _add_en_passant(self, c: int):
        """Adds an en passant move into the given column."""
        self._add_move(self._pos.r + self._dr, c)



"""
Player class.
"""

# =============================================================================

from typing import Union

from alicechess.position import Move
from alicechess.utils import Color, PromoteType

# =============================================================================

__all__ = (
    "Player",
    "HumanPlayer",
)

# =============================================================================


class _PlayerBase:
    """Base class for a player.."""

    is_human = False

    def __init__(self, color: Color):
        """Initializes a player."""
        self._color = color

    @property
    def color(self) -> Color:
        """The player color."""
        return self._color


class Player(_PlayerBase):
    """Represents a player.

    Properties:
        color (Color): The player color.
        is_human (bool): Whether this player is human.

    Methods:
        make_move(game_state) -> Move
            Called when this player should make their next move.
        promote(game_state) -> PromoteType
            Called when this player can promote a pawn.
    """

    def make_move(self, game_state: "GameState") -> Move:
        """Called when this player should make their next move.

        Args:
            game_state (GameState): The current game state.

        Returns:
            Move: The move to make.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}: method `make_move()` not implemented"
        )

    def promote(self, game_state: "GameState") -> PromoteType:
        """Called when this player can promote a pawn.

        The move being made is accessible through `game_state.move`.

        Args:
            game_state (GameState): The current game state.

        Returns:
            PromoteType: The piece type to promote to.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}: method `promote()` not implemented"
        )


class HumanPlayer(_PlayerBase):
    """Represents a human player."""

    is_human = True


# =============================================================================

AnyPlayer = Union[Player, HumanPlayer]



"""
Game class.
"""

# =============================================================================

from typing import Type

from alicechess.game_state import GameState
from alicechess.player import AnyPlayer, _PlayerBase

# =============================================================================

__all__ = ("Game",)

# =============================================================================


class Game:
    """Represents a game of Alice Chess."""

    def __init__(self, *, white: Type[AnyPlayer], black: Type[AnyPlayer]):
        """Initializes a Game with the given players.

        The players must be subclasses of `Player` (not instances).

        Args:
            white (Type[Player]): The player for white.
            black (Type[Player]): The player for black.
        """

        if not isinstance(white, type):
            raise TypeError("`white` must be a class (not an instance)")
        if not isinstance(black, type):
            raise TypeError("`black` must be a class (not an instance)")
        if not issubclass(white, _PlayerBase):
            raise TypeError("`white` must be a Player subclass")
        if not issubclass(black, _PlayerBase):
            raise TypeError("`black` must be a Player subclass")

        self._white = white
        self._black = black

    def new(self) -> GameState:
        """Returns a new game."""
        return GameState.new(white=self._white, black=self._black)

    def start_window(self, **kwargs):
        """Starts the game in a window.

        Any keyword arguments will be passed to the `Window`
        constructor.
        """
        # import here so that games that don't use the window don't need
        # to be run in a virtual environment (requires Pillow)
        # pylint: disable=import-outside-toplevel
        from alicechess.window import Window

        game_state = self.new()
        window = Window(game_state, **kwargs)
        window.run()



"""
MovesCalculator class.
"""

# =============================================================================

from functools import partial
from typing import Dict, Iterator, List, Optional, Sequence, Tuple

from alicechess.pieces import King, Pawn, Piece, Rook
from alicechess.position import BoardPosition, Position
from alicechess.utils import Color, PieceType, check_brc, check_brc_bool

# =============================================================================

BoardDict = Dict[BoardPosition, Piece]

# =============================================================================

STRAIGHT_DIRS = (
    (0, -1),  # left
    (-1, 0),  # up
    (0, 1),  # right
    (1, 0),  # down
)

DIAGONAL_DIRS = (
    (-1, -1),  # up left
    (-1, 1),  # up right
    (1, 1),  # down right
    (1, -1),  # down left
)

ALL_DIRS = STRAIGHT_DIRS + DIAGONAL_DIRS

KNIGHT_MOVES = (
    (-1, -2),  # left up
    (-2, -1),  # up left
    (-2, 1),  # up right
    (-1, 2),  # right up
    (1, 2),  # right down
    (2, 1),  # down right
    (2, -1),  # down left
    (1, -2),  # left down
)


def _yield_pos(directions, pr, pc, steps=None) -> Iterator[Tuple[int, int]]:
    for dr, dc in directions:
        r = pr
        c = pc
        i = 0
        while steps is None or i < steps:
            r += dr
            c += dc
            if not check_brc_bool(r=r, c=c):
                break
            stop = yield r, c
            if stop:
                # last yield for send
                yield None
                break
            i += 1


yield_straight_pos = partial(_yield_pos, STRAIGHT_DIRS)
yield_diagonal_pos = partial(_yield_pos, DIAGONAL_DIRS)
yield_all_dir_pos = partial(_yield_pos, ALL_DIRS)
yield_knight_pos = partial(_yield_pos, KNIGHT_MOVES, steps=1)


POS_GEN_FUNCS = {
    PieceType.KING: partial(yield_all_dir_pos, steps=1),
    PieceType.QUEEN: yield_all_dir_pos,
    PieceType.ROOK: yield_straight_pos,
    PieceType.KNIGHT: yield_knight_pos,
    PieceType.BISHOP: yield_diagonal_pos,
}

# =============================================================================


def _get(board: BoardDict, bn: int, r: int, c: int):
    return board.get((bn, r, c), None)


# =============================================================================


def _yield_threatened_by(
    board: BoardDict, by_color: Color, bn: int, r: int, c: int
) -> Iterator[Piece]:
    """Yields the pieces that the given position is threatened by."""
    check_brc(bn, r, c)

    piece_at_pos = _get(board, bn, r, c)
    if piece_at_pos is not None:
        if piece_at_pos.color is by_color:
            # own piece can't be threatened
            return

    # check knight positions
    for tr, tc in yield_knight_pos(r, c):
        piece = _get(board, bn, tr, tc)
        if piece is None or piece.color is not by_color:
            continue
        if piece.type is PieceType.KNIGHT:
            yield piece

    # check pawn positions
    if by_color is Color.WHITE:
        # threaten from bottom
        dr = 1
    else:
        # threaten from top
        dr = -1
    for dc in (-1, 1):
        piece = _get(board, bn, r + dr, c + dc)
        if piece is None or piece.color is not by_color:
            continue
        if piece.type is PieceType.PAWN:
            yield piece

    # check king positions
    for tr, tc in POS_GEN_FUNCS[PieceType.KING](r, c):
        piece = _get(board, bn, tr, tc)
        if piece is None or piece.color is not by_color:
            continue
        if piece.type is PieceType.KING:
            yield piece

    # check straights and diagonals
    for pos_gen_func, piece_types in (
        (yield_straight_pos, (PieceType.QUEEN, PieceType.ROOK)),
        (yield_diagonal_pos, (PieceType.QUEEN, PieceType.BISHOP)),
    ):
        pos_gen = pos_gen_func(r, c)
        for tr, tc in pos_gen:
            piece = _get(board, bn, tr, tc)
            if piece is None:
                continue
            if piece.color is by_color and piece.type in piece_types:
                yield piece
            # done with this direction
            pos_gen.send(True)


def _is_threatened(
    board: BoardDict, by_color: Color, bn: int, r: int, c: int
) -> bool:
    try:
        next(_yield_threatened_by(board, by_color, bn, r, c))
        return True
    except StopIteration:
        # nothing yielded
        return False


# =============================================================================


class MovesCalculator:
    """Calculates the possible moves for a given board state.

    Methods:
        num_moves(color) -> int
            Returns the number of moves for the given color.
        is_in_check(color) -> bool
            Returns whether the given color is in check.
    """

    def __init__(
        self,
        board: BoardDict,
        kings: Tuple[King],
        unmoved_rooks: Tuple[Sequence[Rook]],
        en_passant_pawn: Optional[Pawn] = None,
    ):
        self._board = board
        self._kings = tuple(kings)
        unmoved_rooks = tuple(map(tuple, unmoved_rooks))

        pawns: Tuple[List[Pawn]] = ([], [])
        for piece in self._board.values():
            if piece.type is PieceType.PAWN:
                pawns[piece.color.value].append(piece)

        self._in_check = tuple(
            _is_threatened(self._board, king.color.other(), *king.pos)
            for king in self._kings
        )
        if self._in_check == (True, True):
            raise ValueError("Both kings are in check")
        self._num_moves = [0, 0]

        # calculate all the possible moves
        for piece in self._board.values():
            self._calc_possible(piece)

        # check for castling
        for king, in_check, rooks in zip(
            self._kings, self._in_check, unmoved_rooks
        ):
            if king.has_moved:
                continue
            if in_check:
                continue
            if len(rooks) == 0:
                continue
            enemy_color = king.color.other()
            bn, kr, kc = king.pos
            for rook in rooks:
                rc = rook.pos.c
                if rc < kc:
                    # left rook: king will move to the left
                    dc = -1
                else:  # kc < rc
                    # right rook: king will move to the right
                    dc = 1
                valid_castle = True
                # check that all the spaces in the middle are empty on
                # both boards
                c = kc + dc
                while c != rc:
                    if (bn, kr, c) in self._board:
                        # there's a piece there
                        valid_castle = False
                        break
                    if (1 - bn, kr, c) in self._board:
                        # there's a piece there on the other board
                        valid_castle = False
                        break
                    c += dc
                if not valid_castle:
                    continue
                # check that the spaces the king moves through are not
                # threatened on this board
                c = kc
                for _ in range(2):
                    c += dc
                    if _is_threatened(self._board, enemy_color, bn, kr, c):
                        valid_castle = False
                        break
                if not valid_castle:
                    continue
                # check that the king does not teleport into check on
                # the other board
                king_c = kc + dc + dc
                if _is_threatened(
                    self._board, enemy_color, 1 - bn, kr, king_c
                ):
                    continue
                # king moves two spaces
                king._add_castle(rook_c=kc + dc, king_c=kc + dc + dc)

        # check for en passant
        if en_passant_pawn is not None:
            if en_passant_pawn.type is not PieceType.PAWN:
                raise ValueError("Given en passant piece is not a pawn")
            pbn, pr, pc = en_passant_pawn.pos
            if pbn != 1:
                raise ValueError("Given en passant pawn is not on Board B")
            # see if any of the enemy's pawns can perform en passant
            for pawn in pawns[en_passant_pawn.color.other().value]:
                bn, r, c = pawn.pos
                if bn != 1:
                    # cannot capture a piece on the other board
                    continue
                if not (r == pr and abs(c - pc) == 1):
                    # pawn is not adjacent to advanced pawn
                    continue
                capture_r = r + pawn.dr
                capture_pos = (capture_r, pc)
                if self._move_in_check(
                    pawn.pos, capture_pos, en_passant=(pr, pc)
                ):
                    continue
                pawn._add_en_passant(pc)

        for piece in self._board.values():
            piece._mark_threatened(False)
            self._num_moves[piece.color.value] += piece.num_moves
        for piece in self._board.values():
            for move in piece.yield_moves():
                threatened_piece = self._get(*move.capture_pos)
                if threatened_piece is not None:
                    threatened_piece._mark_threatened(True)

    def _get(self, bn: int, r: int, c: int) -> Optional[Piece]:
        return _get(self._board, bn, r, c)

    def num_moves(self, color: Color) -> int:
        """Returns the number of moves for the given color."""
        return self._num_moves[color.value]

    def is_in_check(self, color: Color) -> bool:
        """Returns whether the given color is in check."""
        return self._in_check[color.value]

    def _move_in_check(
        self,
        pos: BoardPosition,
        target: Position,
        en_passant: Optional[Position] = None,
    ):
        pos = BoardPosition.of(pos)
        tr, tc = target
        here_target_pos = (pos.bn, tr, tc)
        there_target_pos = (1 - pos.bn, tr, tc)

        moving = self._get(*pos)
        if moving is None:
            raise ValueError(f"No piece at position {pos} {pos.pos}")

        # make temp copy board (no need to copy pieces because not
        # changing them)
        board = dict(self._board)

        # check move validity
        if en_passant is None:
            captured = _get(board, *here_target_pos)
            if captured is not None:
                if captured.type is PieceType.KING:
                    raise ValueError("Move captures a king")
                if captured.color is moving.color:
                    raise ValueError("Move captures own piece")
        else:
            # capture the en passant piece
            captured = board.pop((pos.bn, *en_passant), None)
            if captured is None:
                raise ValueError("Given en passant position is blank")
            if captured.type is not PieceType.PAWN:
                raise ValueError("Cannot perform en passant on a non-pawn")
            if captured.color is moving.color:
                raise ValueError("En passant captures own piece")
        other_board = _get(board, *there_target_pos)
        if other_board is not None:
            raise ValueError("Move squashes a piece on the other board")

        enemy_color = moving.color.other()
        king = self._kings[moving.color.value]
        moving_king = moving is king

        # make sure move is valid on current board
        board[here_target_pos] = board.pop(pos).move_to(here_target_pos)
        if moving_king:
            check_pos = here_target_pos
        else:
            check_pos = king.pos
        if _is_threatened(board, enemy_color, *check_pos):
            return True
        # make sure move is valid after teleporting
        board[there_target_pos] = board.pop(here_target_pos).move_to(
            there_target_pos
        )
        if moving_king:
            check_pos = there_target_pos
        else:
            check_pos = king.pos
        if _is_threatened(board, enemy_color, *check_pos):
            return True

        return False

    def _calc_possible_pawn(self, piece: Pawn):
        bn, pr, pc = piece.pos
        moves = set()

        def check_pos(r, c):
            if not check_brc_bool(r=r, c=c):
                return
            this_board = self._get(bn, r, c)
            if this_board is not None:
                return
            other_board = self._get(1 - bn, r, c)
            if other_board is not None:
                return
            if self._move_in_check(piece.pos, (r, c)):
                return
            moves.add(Position(r, c))

        r = pr + piece.dr

        if not check_brc_bool(r=r):
            # row is out of bounds; the pawn is on the promotion row
            return moves

        # forward
        c = pc
        check_pos(r, c)
        if not piece.has_moved and self._get(bn, r, c) is None:
            # check double forward
            check_pos(r + piece.dr, c)

        # diagonal capture
        for dc in (-1, 1):
            c = pc + dc
            if not check_brc_bool(c=c):
                continue
            front_diag = self._get(bn, r, c)
            if front_diag is None:
                # must have a piece there
                continue
            this_board = self._get(bn, r, c)
            if this_board is not None:
                if this_board.type is PieceType.KING:
                    # can't capture a king
                    continue
                if this_board.color is piece.color:
                    # can't capture your own piece
                    continue
            other_board = self._get(1 - bn, r, c)
            if other_board is not None:
                # can't replace on other board
                continue
            if self._move_in_check(piece.pos, (r, c)):
                continue
            moves.add(Position(r, c))

        return moves

    def _calc_possible(self, piece: Piece):
        if self._get(*piece.pos) is not piece:
            raise ValueError(
                f"Piece {piece.name!r} is not at the proper position "
                f"({piece.pos}) in the board"
            )
        if piece.type is PieceType.PAWN:
            moves = self._calc_possible_pawn(piece)
        else:
            bn, pr, pc = piece.pos
            moves = set()
            pos_gen = POS_GEN_FUNCS[piece.type](pr, pc)
            for r, c in pos_gen:
                this_board = self._get(bn, r, c)
                if this_board is not None:
                    # done with this direction
                    pos_gen.send(True)
                    if this_board.type is PieceType.KING:
                        # can't capture a king
                        continue
                    if this_board.color is piece.color:
                        # can't capture your own piece
                        continue
                other_board = self._get(1 - bn, r, c)
                if other_board is not None:
                    # can't replace on other board
                    continue
                if self._move_in_check(piece.pos, (r, c)):
                    continue
                moves.add(Position(r, c))
        piece._set_possible_moves(moves)  # pylint: disable=protected-access



"""
Alice Chess.
"""

from alicechess.game import Game
from alicechess.game_state import GameState
from alicechess.player import HumanPlayer, Player
from alicechess.utils import Color, PieceType, PromoteType



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['alicechess', 'alicechess.pieces']

package_data = \
{'': ['*'], 'alicechess': ['pictures/*']}

install_requires = \
['Pillow>=9.5.0,<10.0.0']

setup_kwargs = {
    'name': 'alicechess',
    'version': '2.1.0',
    'description': 'A Python package to play Alice Chess',
    'long_description': '# Alice Chess\n\nThis project allows you to play Alice Chess, a variant of chess.\n\n## Installation\n\nThe package may be installed through `pip`:\n\n```bash\n$ pip install alicechess\n```\n\n## Rules\n\nHere is a [description of the concept and rules][rules].\n\n[rules]: https://www.chessvariants.com/other.dir/alice.html\n\nNotable game rules:\n\n- **Castling**: A king and rook may only castle if neither has moved already,\n  the king is not in check, all squares between them are vacant on both boards,\n  and the king does not move through check or into check. After the castle, both\n  pieces will teleport to the other board.\n- **En passant**: A pawn may capture another pawn through en passant if your\n  pawn is on board B and the enemy pawn advances two spaces, teleporting to the\n  space right next to yours on board B. (This results in a situation that looks\n  like regular en passant.) Note that due to teleporting to the other board\n  after each move, this can only be achieved by a pawn that _does not_ advance\n  two squares on its first move.\n- **Fifty move rule**: If both players have made 50 moves each where no piece\n  has been captured or no pawn moved, then a player may claim a draw. However,\n  to simplify this case, the game will be automatically ended with a draw\n  (rather than allowing a player to claim a draw), although this is not the\n  official rule. This therefore overshadows the 75-move rule, where a draw is\n  automatically applied after 75 moves by both players with no captures or pawn\n  movements.\n\n## How to play\n\nTo start a game between two human players, you can easily just run the package\non the command line:\n\n```bash\n$ python -m alicechess\n```\n\nA window will come up where the game can be played.\n\n## Playing against bots\n\nTo play a game against a bot or between bots, you must write your own script.\nYou should initialize a `Game` object with the appropriate players, then call\neither the `start()` or `start_window()` method. To write your own bot, see the\n[Writing a bot](#writing-a-bot) section.\n\nHere is an example:\n\n```python\n"""\nPlays a game between a human and a bot that plays randomly.\n"""\n\nfrom alicechess import Game, HumanPlayer\nfrom alicechess.bots import RandomPlayer\n\nif __name__ == "__main__":\n    Game(white=HumanPlayer, black=RandomPlayer).start_window()\n```\n\nNote that the class names (not instances) are passed to the `Game` constructor.\n\nThe `start_window()` method will, as implied, start an interactive window where\nthe game can be played. However, you can also opt to use the `start()` method\ninstead, which will return the first `GameState` of the game, and then use\nanother way to ask the user for input and play the game; for instance, you could\nmake the game entirely textual with user input provided with the keyboard. See\nthe [API Documentation][docs] for more information.\n\nIn the interactive window, there is a 3 second delay for non-human player moves,\nto simulate realism. This can be changed by passing a value for\n`non_human_player_delay` to the `start_window()` method.\n\nIt is also possible for two bots to play against each other.\n\n### Writing a bot\n\nThe code is factored in a way to make it very easy for you to code your own bots\nto play against. Simply extend the `Player` class and implement the two abstract\nmethods for making a move and promoting a pawn. This class (not an instance) can\nthen be passed into the `Game` constructor to start a game. See the\n[API Documentation][docs] for more information.\n\nHere is an example:\n\n```python\n"""\nPlays a game between a human and a bot (that I wrote).\n"""\n\nfrom alicechess import Game, HumanPlayer, Player, PromoteType\n\nclass Bot(Player):\n    """A very good bot that I wrote."""\n\n    def make_move(self, game_state):\n        for piece in game_state.yield_player_pieces():\n            for move in piece.yield_moves():\n                return move\n\n    def promote(self, game_state):\n        return PromoteType.QUEEN\n\nif __name__ == "__main__":\n    Game(white=HumanPlayer, black=Bot).start_window()\n```\n\n[docs]: https://github.com/josephlou5/alicechess/blob/main/Documentation.md\n\n## Credit\n\nThank you to Artyom Lisitsyn for inspiring me to pursue this project and to\nTrung Phan for being my chess consultant and answering all my questions on rules\nand technicalities.\n',
    'author': 'Joseph Lou',
    'author_email': 'joseph.d.lou@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/josephlou5/alicechess',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



"""
King piece.
"""

# =============================================================================

from alicechess.pieces.piece import Piece
from alicechess.utils import Color, PieceType

# =============================================================================


class King(Piece):
    """King piece.

    Properties:
        can_castle_left (bool): Whether the king can castle to the left
            (queenside).
        left_rook_col (Optional[int]): The column the left rook castles
            to, if the king can castle with it.
        can_castle_right (bool): Whether the king can castle to the
            right (kingside).
        right_rook_col (Optional[int]): The column the right rook
            castles to, if the king can castle with it.
    """

    _type = PieceType.KING

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._left_rook_col = None
        self._right_rook_col = None

        if self._color is Color.WHITE:
            start_row = 7
        else:
            start_row = 0
        # king is always on the right (for this particular board setup)
        self._start_pos = (0, start_row, 4)

    def _check_start_position(self) -> bool:
        return self._pos == self._start_pos

    @property
    def can_castle_left(self) -> bool:
        """Whether the king can castle to the left (queenside)."""
        return self._left_rook_col is not None

    @property
    def left_rook_col(self):
        """The column the left rook castles to, if the king can castle
        with it.
        """
        return self._left_rook_col

    @property
    def can_castle_right(self) -> bool:
        """Whether the king can castle to the right (kingside)."""
        return self._right_rook_col is not None

    @property
    def right_rook_col(self):
        """The column the right rook castles to, if the king can castle
        with it.
        """
        return self._right_rook_col

    def _add_castle(self, rook_c: int, king_c: int):
        """Adds a castle to the given column."""
        if self._moves is None:
            raise RuntimeError("King has not calculated its possible moves")
        # since the board is constructed with white on the bottom,
        # technically the castle columns are always fixed
        _, kr, kc = self._pos
        if king_c < kc:
            # left castle
            self._left_rook_col = rook_c
        else:
            # right castle
            self._right_rook_col = rook_c
        # add the move
        self._add_move(kr, king_c)
